use std::convert::TryInto;
use std::net::{IpAddr, Ipv4Addr, Ipv6Addr, UdpSocket};
use std::str::FromStr;

use log::*;

use doge_dns::{Request, Response};
use crate::GenericTransport;

use crate::{Transport, Error};


/// The **UDP transport**, which sends DNS wire data inside a UDP datagram.
///
/// # References
///
/// - [RFC 1035 §4.2.1](https://tools.ietf.org/html/rfc1035) — Domain Names,
///   Implementation and Specification (November 1987)
pub struct UdpTransport {
    addr: String,
    port: u16,
}

impl UdpTransport {

    /// Creates a new UDP transport that connects to the given host.
    pub fn new(addr: GenericTransport) -> Self {
        if addr.port_num != 0 {
            Self {
            addr : addr.address,
            port : addr.port_num,
            }
        } else {
            Self {
                addr : addr.address,
                port : 53
            }
        }
    }
}


impl Transport for UdpTransport {
    fn send(&self, request: &Request) -> Result<Response, Error> {
        info!("Opening UDP socket");
        let ip_addr = IpAddr::from_str(&self.addr).expect("Failed to parse Ip addr");

        let socket: UdpSocket = match ip_addr.is_ipv4() {
            true => UdpSocket::bind((Ipv4Addr::UNSPECIFIED, 0)),
            false => UdpSocket::bind((Ipv6Addr::UNSPECIFIED, 0)),
        }.expect("Unknown ip format");

        socket.connect( (ip_addr, self.port))?;

        debug!("Opened");

        let bytes_to_send = request.to_bytes().expect("failed to serialise request");

        info!("Sending {} bytes of data to {} over UDP", bytes_to_send.len(), self.addr);
        let written_len = socket.send(&bytes_to_send)?;
        debug!("Wrote {} bytes", written_len);

        info!("Waiting to receive...");
        let mut buf = vec![0; 4096];
        let received_len = socket.recv(&mut buf)?;

        info!("Received {} bytes of data", received_len);
        let response = Response::from_bytes(&buf[.. received_len])?;
        Ok(response)
    }
}
